@@ -1,19 +1,10 @@
 # =====================================================================================================================
 # VERSION = (0, 0, 1)   # use import EXACT_OBJECTS! not *
-#   from .main import *                 # INcorerct
-#   from .main import EXACT_OBJECTS     # CORERCT
-
-
-# =====================================================================================================================
-# from .base import (
-#     AlertBase,
-# )
-
-
-# =====================================================================================================================
-from .sys_git import (
-    Git,
-)
+#   from .main import *                 # INcorrect
+#   from .main import EXACT_OBJECTS     # CORRECT
+# VERSION = (0, 0, 2)   # del blank lines
+# VERSION = (0, 0, 3)   # separate all types/exx into static.py!
+# VERSION = (0, 0, 4)   # DELETE ALL IMPORTS! dont use it! CIRCULAR IMPORTS
 
 
 # =====================================================================================================================
